from collections import namedtuple
from itertools import chain
from typing import NamedTuple, Mapping, List, Tuple

import funcy as fn
from bidict import bidict
from parsimonious import Grammar, NodeVisitor
from toposort import toposort

from aiger import aig

_Symbol = namedtuple('Symbol', ['kind', 'index', 'name'])
_SymbolTable = namedtuple('SymbolTable', ['inputs', 'outputs', 'latches'])

AAG_GRAMMAR = Grammar(u'''
aag = header ios latches ios gates symbols comments?
header = "aag" _ id _ id _ id _ id _ id EOL

ios = io*
io = id EOL

latch_or_gate = id _ id _ id EOL?

latches = (latch / latch_or_gate)*
latch = id _ id EOL

gates = latch_or_gate*

symbols = symbol*
symbol =  symbol_kind id _ symbol_name EOL
symbol_kind = ("i" / "o" / "l")
symbol_name = (~r".")+

comments = "c" EOL comment+
comment = (~r".")* EOL?

_ = ~r" "+
id = ~r"\\d"+
EOL = "\\n"
''')


class AAGVisitor(NodeVisitor):
    def generic_visit(self, _, children):
        return children

    def visit_id(self, node, children):
        return int(node.text)

    def visit_header(self, _, children):
        return Header(*map(int, children[2::2]))

    def visit_io(self, _, children):
        return int(children[::2][0])

    def visit_latches(self, _, children):
        return list(fn.pluck(0, children))

    def visit_latch_or_gate(self, _, children):
        return list(map(int, children[::2]))

    visit_latch = visit_latch_or_gate

    def visit_aag(self, _, children):
        header, ios1, lgs1, ios2, lgs2, symbols, comments = children
        ios, lgs = ios1 + ios2, lgs1 + lgs2
        assert len(ios) == header.num_inputs + header.num_outputs
        inputs, outputs = ios[:header.num_inputs], ios[header.num_inputs:]
        assert len(lgs) == header.num_ands + header.num_latches

        latches, gates = lgs[:header.num_latches], lgs[header.num_latches:]

        # TODO: need to allow for inputs, outputs, latches not in
        # symbol table.
        inputs = {
            symbols.inputs.inv.get(idx, f'i{idx}'): i
            for idx, i in enumerate(inputs)
        }
        outputs = {
            symbols.outputs.inv.get(idx, f'o{idx}'): i
            for idx, i in enumerate(outputs)
        }

        latches = {
            symbols.latches.inv.get(idx, f'l{idx}'): tuple(i)
            for idx, i in enumerate(latches)
        }
        latches = fn.walk_values(lambda l: (l + (0, ))[:3], latches)

        if len(comments) > 0:
            assert comments[0].startswith('c\n')
            comments[0] = comments[0][2:]
        return AAG(
            inputs=inputs,
            outputs=outputs,
            latches=fn.walk_values(tuple, latches),
            gates=fn.lmap(tuple, gates),
            comments=tuple(comments))

    def visit_symbols(self, node, children):
        children = {(k, int(i), n) for k, i, n in children}

        def to_dict(kind):
            return bidict({n: i for k, i, n in children if k == kind})

        return _SymbolTable(to_dict('i'), to_dict('o'), to_dict('l'))

    def visit_symbol(self, node, children):
        return _Symbol(children[0], int(children[1]), children[3])

    def node_text(self, node, _):
        return node.text

    visit_symbol_kind = node_text
    visit_symbol_name = node_text
    visit_comments = node_text


class Header(NamedTuple):
    max_var_index: int
    num_inputs: int
    num_latches: int
    num_outputs: int
    num_ands: int


def _to_idx(lit):
    """AAG format uses least significant bit to encode an inverter.
    The index is thus the interal literal shifted by one bit."""
    return lit >> 1


def _polarity(i):
    return aig.Inverter if i & 1 == 1 else lambda x: x


class AAG(NamedTuple):
    inputs: Mapping[str, int]
    latches: Mapping[str, Tuple[int]]
    outputs: Mapping[str, int]
    gates: List[Tuple[int]]
    comments: Tuple[str]

    @property
    def header(self):
        literals = chain(
            self.inputs.values(),
            self.outputs.values(),
            fn.pluck(0, self.gates), fn.pluck(0, self.latches.values())
        )
        max_idx = max(map(_to_idx, literals), default=0)
        return Header(max_idx, *map(len, self[:-1]))

    def __repr__(self):
        if self.inputs:
            input_names, input_lits = zip(*list(self.inputs.items()))
        if self.outputs:
            output_names, output_lits = zip(*list(self.outputs.items()))
        if self.latches:
            latch_names, latch_lits = zip(*list(self.latches.items()))

        out = f"aag " + " ".join(map(str, self.header)) + '\n'
        if self.inputs:
            out += '\n'.join(map(str, input_lits)) + '\n'
        if self.latches:
            out += '\n'.join([' '.join(map(str, xs))
                              for xs in latch_lits]) + '\n'
        if self.outputs:
            out += '\n'.join(map(str, output_lits)) + '\n'

        if self.gates:
            out += '\n'.join([' '.join(map(str, xs))
                              for xs in self.gates]) + '\n'
        if self.inputs:
            out += '\n'.join(
                f"i{idx} {name}" for idx, name in enumerate(input_names)
            ) + '\n'
        if self.outputs:
            out += '\n'.join(
                f"o{idx} {name}" for idx, name in enumerate(output_names)
            ) + '\n'

        if self.latches:
            out += '\n'.join(
                f"l{idx} {name}" for idx, name in enumerate(latch_names)
            ) + '\n'

        if self.comments:
            out += 'c\n' + '\n'.join(self.comments)
            if out[-1] != '\n':
                out += '\n'
        return out

    def _to_aig(self):
        gate_order, latch_order = self.eval_order_and_gate_lookup

        lookup = fn.merge(
            {0: aig.ConstFalse()},
            {_to_idx(l): aig.Input(n) for n, l in self.inputs.items()},
            {
                _to_idx(l): aig.LatchIn(n)
                for n, (l, _, init) in self.latches.items()
            },
        )
        latches = set()
        and_dependencies = {i: (l, r) for i, l, r in self.gates}
        for gate in fn.cat(gate_order):
            if _to_idx(gate) in lookup:
                continue

            inputs = and_dependencies[gate]
            sources = [_polarity(i)(lookup[_to_idx(i)]) for i in inputs]
            lookup[_to_idx(gate)] = aig.AndGate(*sources)

        latch_dependencies = {
            i: (n, dep) for n, (i, dep, _) in self.latches.items()
        }
        for gate in fn.cat(latch_order):
            assert _to_idx(gate) in lookup
            if not isinstance(lookup[_to_idx(gate)], aig.LatchIn):
                continue

            name, dep = latch_dependencies[gate]
            source = _polarity(dep)(lookup[_to_idx(dep)])
            latches.add((name, source))

        def get_output(v):
            idx = _to_idx(v)
            return _polarity(v)(lookup[idx])

        top_level = ((k, get_output(v)) for k, v in self.outputs.items())
        return aig.AIG(
            inputs=frozenset(self.inputs),
            node_map=frozenset(top_level),
            latch_map=frozenset(latches),
            latch2init=frozenset(
                (n, bool(init)) for n, (_, _, init) in self.latches.items()
            ),
            comments=self.comments
        )

    @property
    def eval_order_and_gate_lookup(self):
        deps = fn.merge(
            {a & -2: {b & -2, c & -2} for a, b, c in self.gates},
            {a & -2: set() for a in self.inputs.values()},
            {a & -2: set() for a, _, _ in self.latches.values()},  # LatchIn
        )
        latch_deps = {a & -2: {b & -2} for a, b, _ in self.latches.values()}
        return list(toposort(deps)), list(toposort(latch_deps))


def aig2aag(circ) -> AAG:
    aag, max_idx, lit_map = _to_aag(
        circ.cones | circ.latch_cones,
        AAG({}, {}, {}, [], circ.comments),
    )

    # Check that all inputs have a lit.
    for name in filter(lambda x: x not in aag.inputs, circ.inputs):
        aag.inputs[name] = lit_map[name] = 2 * max_idx
        max_idx += 1

    # Update cone maps.
    aag.outputs.update(
        {k: lit_map[cone] for k, cone in circ.node_map.items()}
    )
    latch2init = dict(circ.latch2init)
    for name, cone in circ.latch_map:
        latch = aig.LatchIn(name)
        if latch not in lit_map:
            lit = lit_map[latch] = 2 * max_idx
            max_idx += 1
        else:
            lit = lit_map[latch]

        init = int(latch2init[name])
        ilit = lit_map[cone]
        aag.latches[name] = lit, ilit, init

    return aag


def _to_aag(gates, aag: AAG = None, *, max_idx=1, lit_map=None):
    if lit_map is None:
        lit_map = {}

    if not gates:
        return aag, max_idx, lit_map

    # Recurse to update get aag for subtrees.
    for c in fn.mapcat(lambda g: g.children, gates):
        if c in lit_map:
            continue
        aag, max_idx, lit_map = _to_aag(
            [c], aag, max_idx=max_idx, lit_map=lit_map
        )

    # Update aag with current level.
    for gate in gates:
        if gate in lit_map:
            continue

        if isinstance(gate, aig.Inverter):
            input_lit = lit_map[gate.input]
            lit_map[gate] = (input_lit & -2) | (1 ^ (input_lit & 1))
            continue
        elif isinstance(gate, aig.ConstFalse):
            lit_map[gate] = 0
            continue

        # Must be And, Latch, or Input
        lit_map[gate] = 2 * max_idx
        max_idx += 1
        if isinstance(gate, aig.AndGate):
            encoded = tuple(map(lit_map.get, (gate, gate.left, gate.right)))
            aag.gates.append(encoded)

        elif isinstance(gate, aig.Input):
            aag.inputs[gate.name] = lit_map[gate]
        else:
            assert isinstance(gate, aig.LatchIn)

    return aag, max_idx, lit_map


def parse(aag_str: str, rule: str = "aag", to_aig=True):
    aag = AAGVisitor().visit(AAG_GRAMMAR[rule].parse(aag_str))
    return aag._to_aig() if to_aig else aag


def load(path: str, rule: str = "aag", to_aig=True):
    with open(path, 'r') as f:
        return parse(''.join(f.readlines()), to_aig=to_aig)



import hypothesis.strategies as st
from hypothesis import given

from aiger.expr import atom, ite


@given(st.data())
def test_expr_and(data):
    x, y = atom('x'), atom('y')
    expr = x & y

    vals = {f'{i}': data.draw(st.booleans()) for i in expr.inputs}
    assert (x(vals) and y(vals)) == expr(vals)


@given(st.data())
def test_expr_and2(data):
    x = atom('x')
    expr = x & x

    vals = {f'{i}': data.draw(st.booleans()) for i in expr.inputs}
    assert x(vals) == expr(vals)


@given(st.data())
def test_expr_or(data):
    x, y = atom('x'), atom('y')
    expr = x | y

    vals = {f'{i}': data.draw(st.booleans()) for i in expr.inputs}
    assert (x(vals) or y(vals)) == expr(vals)


@given(st.data())
def test_expr_xor(data):
    x, y = atom('x'), atom('y')
    expr = x ^ y

    vals = {f'{i}': data.draw(st.booleans()) for i in expr.inputs}
    assert (x(vals) ^ y(vals)) == expr(vals)


@given(st.data())
def test_expr_eq(data):
    x, y = atom('x'), atom('y')
    expr = x == y

    vals = {f'{i}': data.draw(st.booleans()) for i in expr.inputs}
    assert (x(vals) == y(vals)) == expr(vals)


@given(st.data())
def test_expr_implies(data):
    x, y = atom('x'), atom('y')
    expr = x.implies(y)

    vals = {f'{i}': data.draw(st.booleans()) for i in expr.inputs}
    assert ((not x(vals)) or y(vals)) == expr(vals)


@given(st.data())
def test_expr_invert(data):
    x = atom('x')
    expr = ~x

    vals = {f'{i}': data.draw(st.booleans()) for i in expr.inputs}
    assert x(vals) == (not expr(vals))


@given(st.data())
def test_expr_ite(data):
    x, y, z = map(atom, 'xyz')
    expr = ite(x, y, z)

    vals = {f'{i}': data.draw(st.booleans()) for i in expr.inputs}
    assert expr(vals) == (vals['y'] if vals['x'] else vals['z'])



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['aiger']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.1,<20.0',
 'bidict>=0.18.0,<0.19.0',
 'funcy>=1.12,<2.0',
 'parsimonious>=0.8.1,<0.9.0',
 'pyrsistent>=0.15.4,<0.16.0',
 'toposort>=1.5,<2.0']

setup_kwargs = {
    'name': 'py-aiger',
    'version': '3.4.0',
    'description': 'A python library for manipulating sequential and-inverter gates.',
    'long_description': '<figure>\n  <img src="assets/logo_text.svg" alt="py-aiger logo" width=300px>\n  <figcaption>\n      pyAiger: A python library for manipulating sequential and\n      combinatorial circuits.\n  </figcaption>\n\n</figure>\n\n\n[![Build Status](https://cloud.drone.io/api/badges/mvcisback/py-aiger/status.svg)](https://cloud.drone.io/mvcisback/py-aiger)\n[![codecov](https://codecov.io/gh/mvcisback/py-aiger/branch/master/graph/badge.svg)](https://codecov.io/gh/mvcisback/py-aiger)\n[![Updates](https://pyup.io/repos/github/mvcisback/py-aiger/shield.svg)](https://pyup.io/repos/github/mvcisback/py-aiger/)\n\n[![PyPI version](https://badge.fury.io/py/py-aiger.svg)](https://badge.fury.io/py/py-aiger)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.1405781.svg)](https://doi.org/10.5281/zenodo.1405781)\n\n<!-- markdown-toc start - Don\'t edit this section. Run M-x markdown-toc-generate-toc again -->\n**Table of Contents**\n\n- [About PyAiger](#about-pyaiger)\n- [Installation](#installation)\n- [Boolean Expression DSL](#boolean-expression-dsl)\n- [Sequential Circuit DSL](#sequential-circuit-dsl)\n    - [Sequential composition](#sequential-composition)\n    - [Parallel composition](#parallel-composition)\n    - [Circuits with Latches/Feedback/Delay](#circuits-with-latchesfeedbackdelay)\n    - [Relabeling](#relabeling)\n    - [Evaluation](#evaluation)\n    - [Useful circuits](#useful-circuits)\n- [Extra](#extra)\n- [Ecosystem](#ecosystem)\n- [Related Projects](#related-projects)\n- [Citing](#citing)\n\n<!-- markdown-toc end -->\n\n\n# About PyAiger\n\n1. Q: How is Py-Aiger pronounced? A: Like "pie" + "grrr".\n2. Q: Why python? Aren\'t you worried about performance?! A: No. The goals of this library are ease of use and hackability. \n3. Q: No, I\'m really concerned about performance! A: This library is not suited to implement logic solvers. For everything else, such as the creation and manipulation of circuits with many thousands of gates in between solver calls, the library is really fast enough.\n4. Q: Where does the name come from? A: <a href="http://fmv.jku.at/aiger/">Aiger</a> is a popular circuit format. The format is used in <a href="http://fmv.jku.at/hwmcc17/">hardware model checking</a>, <a href="http://www.syntcomp.org/">synthesis</a>, and is supported by <a href="https://github.com/berkeley-abc/abc">ABC</a>. The name is a combination of AIG (standing for <a href="https://en.wikipedia.org/wiki/And-inverter_graph">And-Inverter-Graph</a>) and the austrian mountain <a href="https://en.wikipedia.org/wiki/Eiger">Eiger</a>.\n\n# Installation\n\nIf you just need to use `aiger`, you can just run:\n\n`$ pip install py-aiger`\n\nFor developers, note that this project uses the\n[poetry](https://poetry.eustace.io/) python package/dependency\nmanagement tool. Please familarize yourself with it and then\nrun:\n\n`$ poetry install`\n\n# Boolean Expression DSL\nWhile powerful, when writing combinatorial circuits, the Sequential\nCircuit DSL can be somewhat clumsy. For this common usecase, we have\ndeveloped the Boolean Expression DSL. All circuits generated this way\nhave a single output.\n\n```python\nimport aiger\nx, y, z = aiger.atoms(\'x\', \'y\', \'z\')\nexpr1 = x & y  # circuit with inputs \'x\', \'y\' and 1 output computing x AND y.\nexpr2 = x | y  # logical or.\nexpr3 = x ^ y  # logical xor.\nexpr4 = x == y  # logical ==, xnor.\nexpr5 = x.implies(y)\nexpr6 = ~x  # logical negation.\nexpr7 = aiger.ite(x, y, z)  # if x then y else z.\n\n# Atoms can be constants.\nexpr8 = x & aiger.atom(True)  # Equivilent to just x.\nexpr9 = x & aiger.atom(False)  # Equivilent to const False.\n\n# And you can inspect the AIG if needed.\ncirc = x.aig\n\n# And of course, you can get a BoolExpr from a single output aig.\nexpr10 = aiger.BoolExpr(circ)\n```\n\n\n# Sequential Circuit DSL\n\n```python\nimport aiger\nfrom aiger import utils\n\n# Parser for ascii AIGER format.\naig1 = aiger.load(path_to_aig1_file.aag)\naig2 = aiger.load(path_to_aig2_file.aag)\n```\n\n## Sequential composition\n```python\naig3 = aig1 >> aig2\n```\n\n## Parallel composition\n```python\naig4 = aig1 | aig2\n```\n\n## Circuits with Latches/Feedback/Delay\n```python\n# Connect output y to input x with delay, initialized to True.\n# (Default initialization is False.)\naig5 = aig1.feedback(\n    inputs=[\'x\'],\n    outputs=[\'y\'],\n    initials=[True],\n    keep_outputs=True\n)\n```\n\n## Relabeling\n```python\n# Relabel input \'x\' to \'z\'.\naig1[\'i\', {\'x\': \'z\'}]\n\n# Relabel output \'y\' to \'w\'.\naig1[\'o\', {\'y\': \'w\'}]\n\n# Relabel latches \'l1\' to \'l2\'.\naig1[\'l\', {\'l1\': \'l2\'}]\n```\n\n## Evaluation\n```python\n# Combinatoric evaluation.\naig3(inputs={\'x\':True, \'y\':False})\n\n# Sequential evaluation.\nsim = aig3.simulate({\'x\': 0, \'y\': 0}, \n                    {\'x\': 1, \'y\': 2},\n                    {\'x\': 3, \'y\': 4})\n\n# Simulation Coroutine\nsim = aig3.simulator()  # Coroutine\nnext(sim)  # Initialize\nprint(sim.send({\'x\': 0, \'y\': 0}))\nprint(sim.send({\'x\': 1, \'y\': 2}))\nprint(sim.send({\'x\': 3, \'y\': 4}))\n\n\n# Unroll\naig4 = aig3.unroll(steps=10, init=True)\n```\n\n## Useful circuits\n```python\n# Fix input x to be False.\naig4 = aiger.source({\'x\': False}) >> aig3\n\n# Remove output y. \naig4 = aig3 >> aiger.sink([\'y\'])\n\n# Create duplicate w of output y.\naig4 = aig3 >> aiger.tee({\'y\': [\'y\', \'w\']})\n\n# Make an AND gate.\naiger.and_gate([\'x\', \'y\'], out=\'name\')\n\n# Make an OR gate.\naiger.or_gate([\'x\', \'y\'])  # Default output name is #or_output.\n\n# And outputs.\naig1 >> aiger.and_gate(aig1.outputs) # Default output name is #and_output.\n\n# Or outputs.\naig1 >> aiger.or_gate(inputs=aig1.outputs, output=\'my_output\')\n\n# Flip outputs.\naig1 >> aiger.bit_flipper(inputs=aig1.outputs)\n\n# Flip inputs.\naiger.bit_flipper(inputs=aig1.inputs) >> aig1\n\n# ITE circuit\n# [\'o1\', \'o2\'] = [\'i1\', \'i2\'] if \'test\' Else [\'i3\', \'i4\'] \naiger.common.ite(\'test\', [\'i1\', \'i2\'], [\'i3\', \'i4\'], outputs=[\'o1\', \'o2\'])\n```\n\n# Extra\n```python\neval_order(aig1)  # Returns topological ordering of circuit gates.\n```\n\n# Ecosystem\n\n<figure>\n <img src="assets/pyaiger_eco.svg" alt="py-aiger ecosystem" width="80%">\n  <figcaption>Overview of the pyaiger ecosystem/stack.</figcaption>\n</figure>\n\n### Stable\n- [py-aiger-bv](https://github.com/mvcisback/py-aiger-bv): Extension of pyAiger for manipulating sequential bitvector circuits.\n- [py-aiger-cnf](https://github.com/mvcisback/py-aiger-cnf): BoolExpr to Object representing CNF. Mostly used for interfacing with py-aiger-sat.\n- [py-aiger-past-ltl](https://github.com/mvcisback/py-aiger-past-ltl): Converts Past Linear Temporal Logic to aiger circuits.\n\n### Underdevelopment\n\n- [py-aiger-coins](https://github.com/mvcisback/py-aiger-coins): Library for creating circuits that encode discrete distributions.\n- [py-aiger-sat](https://github.com/mvcisback/py-aiger-sat): Bridge between py-aiger and py-sat.\n- [py-aiger-bdd](https://github.com/mvcisback/py-aiger-bdd): Aiger <-> BDD bridge.\n- [py-aiger-gridworld](https://github.com/mvcisback/py-aiger-gridworld): Create aiger circuits representing gridworlds.\n- [py-aiger-spectral](https://github.com/mvcisback/py-aiger-spectral): A tool for performing (Fourier) Analysis of Boolean Functions.\n- [py-aigar](https://github.com/mvcisback/py-aigar): pyAiger-Analysis: Batteries included tools for analyzing aiger circuits. \n\n# Related Projects\n- [pyAig](https://bitbucket.org/sterin/pyaig): Another python library\n  for working with AIGER circuits.\n\n\n# Citing\n\n```\n@misc{pyAiger,\n  author       = {Marcell Vazquez-Chanlatte},\n  title        = {mvcisback/py-aiger},\n  month        = aug,\n  year         = 2018,\n  doi          = {10.5281/zenodo.1326224},\n  url          = {https://doi.org/10.5281/zenodo.1326224}\n}\n```\n',
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'marcell.vc@eecs.berkeley.edu',
    'url': 'https://github.com/mvcisback/py-aiger',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from functools import reduce
from typing import Tuple, FrozenSet, NamedTuple, Union

import attr
import funcy as fn
from pyrsistent import pmap

from aiger import common as cmn
from aiger import parser


def timed_name(name, time):
    return f"{name}##time_{time}"


class AndGate(NamedTuple):
    left: 'Node'  # TODO: replace with Node once 3.7 lands.
    right: 'Node'

    @property
    def children(self):
        return (self.left, self.right)

    def __hash__(self):
        return id(self)


class Inverter(NamedTuple):
    input: 'Node'

    @property
    def children(self):
        return (self.input, )

    def __hash__(self):
        return id(self)


# Enables filtering for Input via lens library.
@attr.s(frozen=True, slots=True, auto_attribs=True)
class Input:
    name: str

    @property
    def children(self):
        return ()


@attr.s(frozen=True, slots=True, auto_attribs=True)
class LatchIn:
    name: str

    @property
    def children(self):
        return ()


class ConstFalse(NamedTuple):
    @property
    def children(self):
        return ()


def _is_const_true(node):
    return isinstance(node, Inverter) and isinstance(node.input, ConstFalse)


Node = Union[AndGate, ConstFalse, Inverter, Input, LatchIn]


@attr.s(frozen=True, slots=True, auto_attribs=True, repr=False)
class AIG:
    inputs: FrozenSet[str] = frozenset()
    node_map: FrozenSet[Tuple[str, Node]] = attr.ib(
        default=pmap(), converter=pmap
    )
    latch_map: FrozenSet[Tuple[str, Node]] = frozenset()
    latch2init: FrozenSet[Tuple[str, Node]] = frozenset()
    comments: Tuple[str] = ()

    _to_aag = parser.aig2aag

    def __repr__(self):
        return repr(self._to_aag())

    def __getitem__(self, others):
        assert isinstance(others, tuple) and len(others) == 2
        kind, relabels = others
        assert kind in {'i', 'o', 'l'}

        basis = {
            'i': self.inputs, 'o': self.outputs, 'l': self.latches
        }.get(kind)
        relabels = fn.project(relabels, basis)

        if kind == 'o':
            relabels = {k: [v] for k, v in relabels.items()}
            return self >> cmn.tee(relabels)
        elif kind == 'i':
            relabels_ = {v: [k] for k, v in relabels.items()}
            return cmn.tee(relabels_) >> self

        # Latches act like inputs and outputs...
        def fix_keys(mapping):
            return fn.walk_keys(lambda x: relabels.get(x, x), mapping)

        circ = self.evolve(
            latch_map=fix_keys(self.latch_map),
            latch2init=fix_keys(self.latch2init)
        )

        def sub(node):
            if isinstance(node, LatchIn):
                return LatchIn(relabels.get(node.name, node.name))
            return node

        return circ._modify_leafs(sub)

    def evolve(self, **kwargs):
        return attr.evolve(self, **kwargs)

    @property
    def aig(self):
        return self

    @property
    def outputs(self):
        return frozenset(self.node_map.keys())

    @property
    def latches(self):
        return frozenset(fn.pluck(0, self.latch2init))

    @property
    def cones(self):
        return frozenset(self.node_map.values())

    @property
    def latch_cones(self):
        return frozenset(fn.pluck(1, self.latch_map))

    def __rshift__(self, other):
        return seq_compose(self, other)

    def __lshift__(self, other):
        return seq_compose(other, self)

    def __or__(self, other):
        return par_compose(self, other)

    def __call__(self, inputs, latches=None):
        if latches is None:
            latches = dict()
        latchins = fn.merge(dict(self.latch2init), latches)
        # Remove latch inputs not used by self.
        latchins = fn.project(latchins, self.latches)

        # Turn into a combinatorial circuit
        circ, lmap = self.cutlatches(self.latches)
        latchins = fn.walk_keys(lambda n: lmap[n][0], latchins)
        inputs = fn.merge(inputs, latchins)

        circ = cmn.source(inputs) >> circ

        all_outputs = {
            n: _is_const_true(node) for n, node in circ.node_map.items()
        }
        outputs = fn.project(all_outputs, self.outputs)
        latch_outputs = fn.omit(all_outputs, self.outputs)

        # Fix up latch names
        lmap_inv = {k: n for n, (k, _) in lmap.items()}
        latch_outputs = fn.walk_keys(lmap_inv.get, latch_outputs)

        return outputs, latch_outputs

    def simulator(self, latches=None):
        inputs = yield
        while True:
            outputs, latches = self(inputs, latches)
            inputs = yield outputs, latches

    def simulate(self, input_seq, latches=None):
        sim = self.simulator()
        next(sim)
        return [sim.send(inputs) for inputs in input_seq]

    def cutlatches(self, latches, check_postcondition=True):
        l_map = {n: (cmn._fresh(), init) for (n, init) in self.latch2init}

        assert len(
            set(fn.pluck(0, l_map.values())) & (self.inputs | self.outputs)
        ) == 0

        def sub(node):
            if isinstance(node, LatchIn):
                return Input(l_map[node.name][0])
            return node

        circ = self._modify_leafs(sub)
        _cones = {l_map[k][0]: v for k, v in circ.latch_map if k in latches}
        aig = self.evolve(
            node_map=circ.node_map + _cones,
            inputs=self.inputs | {n for n, _ in l_map.values()},
            latch_map={(k, v) for k, v in circ.latch_map if k not in latches},
            latch2init={(k, v) for k, v in self.latch2init if k not in latches}
        )
        return aig, l_map

    def feedback(
        self, inputs, outputs, initials=None, latches=None, keep_outputs=False
    ):
        if latches is None:
            latches = inputs

        if initials is None:
            initials = [False for _ in inputs]

        assert len(inputs) == len(initials) == len(outputs) == len(latches)
        assert len(set(inputs) & self.inputs) != 0
        assert len(set(outputs) & self.outputs) != 0

        in2latch = {iname: lname for iname, lname in zip(inputs, latches)}

        def sub(node):
            if isinstance(node, Input) and node.name in inputs:
                return LatchIn(in2latch[node.name])
            return node

        aig = self._modify_leafs(sub)

        _latch_map, node_map = fn.lsplit(
            lambda x: x[0] in outputs, aig.node_map.items()
        )
        out2latch = {oname: lname for oname, lname in zip(outputs, latches)}
        _latch_map = {(out2latch[k], v) for k, v in _latch_map}
        l2init = frozenset((n, val) for n, val in zip(latches, initials))
        return aig.evolve(
            inputs=aig.inputs - set(inputs),
            node_map=aig.node_map if keep_outputs else pmap(node_map),
            latch_map=aig.latch_map | _latch_map,
            latch2init=aig.latch2init | l2init
        )

    def unroll(self, horizon, *, init=True, omit_latches=True,
               only_last_outputs=False):
        # TODO:
        # - Check for name collisions.
        latches = self.latches
        aag0, l_map = self.cutlatches({l for l in latches})

        def _unroll():
            prev = aag0
            for t in range(1, horizon + 1):
                tmp = prev['i', {k: timed_name(k, t - 1) for k in aag0.inputs}]
                yield tmp['o', {k: timed_name(k, t) for k in aag0.outputs}]

        unrolled = reduce(seq_compose, _unroll())
        if init:
            source = {timed_name(n, 0): init for n, init in l_map.values()}
            unrolled = cmn.source(source) >> unrolled

        if omit_latches:
            latch_names = [timed_name(n, horizon) for n, _ in l_map.values()]
            unrolled = unrolled >> cmn.sink(latch_names)

        if only_last_outputs:
            odrop = fn.lfilter(
                lambda o: int(o.split('##time_')[1]) < horizon,
                unrolled.outputs
            )
            unrolled = unrolled >> cmn.sink(odrop)

        return unrolled

    def write(self, path):
        with open(path, 'w') as f:
            f.write(repr(self))

    def _modify_leafs(self, func):
        @fn.memoize
        def _mod(node):
            if isinstance(node, AndGate):
                left = _mod(node.left)
                right = _mod(node.right)
                if ConstFalse() in (left, right):
                    return ConstFalse()
                elif _is_const_true(left):
                    return right
                elif _is_const_true(right):
                    return left
                else:
                    return node._replace(left=left, right=right)

            elif isinstance(node, Inverter):
                child = _mod(node.input)
                if isinstance(child, Inverter):
                    return child.input
                else:
                    return node._replace(input=child)

            return func(node)

        node_map = ((name, _mod(cone)) for name, cone in self.node_map.items())
        latch_map = ((name, _mod(cone)) for name, cone in self.latch_map)
        return self.evolve(
            node_map=pmap(node_map),
            latch_map=frozenset(latch_map)
        )


def par_compose(aig1, aig2, check_precondition=True):
    assert not aig1.latches & aig2.latches
    assert not aig1.outputs & aig2.outputs

    shared_inputs = aig1.inputs & aig2.inputs
    if shared_inputs:
        relabels1 = {n: cmn._fresh() for n in shared_inputs}
        relabels2 = {n: cmn._fresh() for n in shared_inputs}
        aig1, aig2 = aig1['i', relabels1], aig2['i', relabels2]

    circ = AIG(
        inputs=aig1.inputs | aig2.inputs,
        latch_map=aig1.latch_map | aig2.latch_map,
        latch2init=aig1.latch2init | aig2.latch2init,
        node_map=aig1.node_map + aig2.node_map,
        comments=aig1.comments + aig2.comments
    )

    if shared_inputs:
        for orig in shared_inputs:
            new1, new2 = relabels1[orig], relabels2[orig]
            circ <<= cmn.tee({orig: [new1, new2]})

    return circ


def seq_compose(circ1, circ2, *, input_kind=Input):
    interface = circ1.outputs & circ2.inputs
    assert not (circ1.outputs - interface) & circ2.outputs
    assert not circ1.latches & circ2.latches

    passthrough = {
        k: v for k, v in circ1.node_map.items() if k not in interface
    }

    circ3 = circ2
    for mapping in [circ1.node_map, circ1.latch_map]:
        lookup = dict(mapping)

        def sub(node):
            if isinstance(node, input_kind):
                return lookup.get(node.name, node)
            return node

        circ3 = circ3._modify_leafs(sub)

    return AIG(
        inputs=circ1.inputs | (circ2.inputs - interface),
        latch_map=circ1.latch_map | circ3.latch_map,
        latch2init=circ1.latch2init | circ2.latch2init,
        node_map=circ3.node_map + passthrough,
        comments=circ1.comments + circ2.comments
    )
